/*
 MessegeCollector

 メッセージ収集プログラム

 */

package main

import (
	"fmt"
	"encoding/json"
	"io/ioutil"
	"log"
	"net/http"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	MQTT "github.com/eclipse/paho.mqtt.golang"

	"github.com/labstack/echo"
	"github.com/labstack/echo/middleware"
)

const (
	CONTROL_URL   = "/control"
	HOST              = "localhost"
	PORT              = uint16(4000)
	BROKER_HOST              = "localhost"
	BROKER_PORT              = uint16(1883)
	CMD_QUIT = "quit"
)

// 変数定義
var (
	myserver_root = "./"
	mqtt_client = MQTT.Client(nil)
	http_client = &http.Client{}
	publish_cliet = MQTT.Client(nil)
	actions = make(map[string]Action)
	debug_mode    = false
)

// configファイルの情報を保持する構造体
type Action struct {
	Topic string `json:"topic"`
	Action string `json:"action"`
	Target string `json:"target"`
	Name string `json:"name"`
}
type Config struct {
	Host              string   `json:"host"`
	Port              uint16   `json:"port"`
	Broker_host       string   `json:"broker_host"`
	Broker_port       uint16   `json:"broker_port"`
	Control_url       string   `json:"control_url"`
	Actions []Action `json:actions`
}

// MQTTパブリッシュ処理
func publishMessage(action Action, msg MQTT.Message){
     // 送信準備
	opts := MQTT.NewClientOptions()
	opts.AddBroker("tcp://" + action.Target)
	opts.SetClientID("Message Collector")
	opts.SetCleanSession(false)

	if debug_mode { fmt.Println(opts) }

	client := MQTT.NewClient(opts)
	if token := client.Connect(); token.Wait() && token.Error() != nil {
		fmt.Println("publishMessage: MQTT client setup failed")
		return		
	}
	fmt.Println("publishMessage", msg.Topic(), "->", "tcp://" + action.Target , action.Name)

	// パブリッシュ
	token := client.Publish(action.Name, byte(0), false, msg.Payload())
	token.Wait()
	if token.Error() != nil { fmt.Println(token.Error()) }

	client.Disconnect(250)

}

// ストレージ・サービスへのメッセージ保存
func saveMessage(action Action, msg MQTT.Message) {
     // ファイル名はタイムスタンプにする
     target_url := action.Target + action.Name + "/"  + msg.Topic() + "/" + strconv.FormatInt(time.Now().UnixNano(), 10) 

	fmt.Println("saveMessage", msg.Topic(), "->", target_url)

     // PUTで保存
     	req, err := http.NewRequest("PUT", target_url, strings.NewReader(string(msg.Payload())))
	if err != nil {
		log.Println("http.NewRequest Failed:", err)
	}
	req.Header.Set("Content-Type", "application/x-www-form-urlencoded")

     client := &http.Client{}
     res, err := client.Do(req)
	if err != nil {
		log.Println("client.Do Failed:", err)
	} else {
		defer res.Body.Close()
	}
}

// MQTTメッセージハンドラ
func messageHandler(client MQTT.Client, msg MQTT.Message) {
     if debug_mode {
	     fmt.Println("start messageHandler")
	     fmt.Println(msg)
	     fmt.Printf("RECEIVED TOPIC: %s MESSAGE: %s\n", msg.Topic(), string(msg.Payload()))
	     fmt.Println("finish messageHandler")
     }
     switch actions[msg.Topic()].Action {
     case "SAVE":
     	  saveMessage(actions[msg.Topic()], msg)
     case "PUBLISH":
     	  publishMessage(actions[msg.Topic()], msg)
     }
}

// データ収集処理の開始
func startCollector(conf *Config) MQTT.Client {

     // MQTTクライアントの準備
     opts := MQTT.NewClientOptions()
     broker := "tcp://" + conf.Broker_host + ":" + strconv.FormatInt(int64(conf.Broker_port), 10)
     opts.AddBroker(broker)
     opts.SetClientID("Message Collector")
     opts.SetCleanSession(false)

     client := MQTT.NewClient(opts)
     token := client.Connect()
     if token == nil { return nil }
     token.Wait()
     if token.Error() != nil { return nil }

     topic := make(map[string]byte)
     for _, action := range conf.Actions {
     	// 収集対象のトッピクを設定
     	 topic[action.Topic] = byte(0)
	 // 各トピックに対する処理を設定
	 actions[action.Topic] = action
     }
     
     if debug_mode { fmt.Println(topic) }

     // サブスクライバーを準備してメッセージを待つ
     token = client.SubscribeMultiple(topic, messageHandler)
     token.Wait()

     return client
}

// リクエストハンドラ：バージョン返却
func requestHandler(c echo.Context) error {
	return c.String(http.StatusOK, "IoT Message Collector Ver0.1")
}

// リクエストハンドラ：プログラムコントロール
func controlHandler(c echo.Context) error {
	if debug_mode { fmt.Println("control handler") }
	// コマンド名取得
	cmd := c.FormValue("command")
	// コマンド毎の処理
	switch cmd {
	case CMD_QUIT: // プログラムの終了
		if debug_mode { fmt.Println(CMD_QUIT) }
		if mqtt_client != nil { fmt.Println("MQTT disconnect"); mqtt_client.Disconnect(250)}
		os.Exit(0)
	}
	return c.NoContent(http.StatusOK)
}

// メイン処理
func main() {

	// プログラム配置場所のTOPを取得
	myserver_root = os.Getenv("MYSERVER_ROOT")
	fmt.Println("MYSERVER_ROOT = ", myserver_root)

	// 設定ファイルを読み込む
	prog_dir := filepath.Dir(os.Args[0])
	config_path := filepath.Join(prog_dir, "config.json")
	log.Println(config_path)
	conf := new(Config)
	conf_file, err := ioutil.ReadFile(config_path)
	if err != nil {
		conf_file, err = ioutil.ReadFile("./config.json")
		if err != nil {
			log.Println("Config file read error.")
			os.Exit(1)
		}
	}
	err = json.Unmarshal(conf_file, conf)
	log.Println(err)
	if err != nil {
		log.Println("Config file parse error.")
		os.Exit(1)
	}
	if debug_mode { log.Println("conf = ", *conf) }

	// 設定ファイルに値がない場合デフォルト値を使う
	if conf.Host == "" { conf.Host = HOST }
	if conf.Port == 0 { conf.Port = PORT }
	if conf.Broker_host == "" { conf.Broker_host = BROKER_HOST }
	if conf.Broker_port == 0 { conf.Broker_port = BROKER_PORT }
	if conf.Control_url == "" { conf.Control_url = CONTROL_URL }
	if debug_mode { log.Println("conf = ", *conf) }


	/*** データ収取処理の起動 ***/
	mqtt_client = startCollector(conf)
	if debug_mode { log.Println("mqtt_client = ", mqtt_client) }
	if mqtt_client == nil {
	   fmt.Println("Failed to subscribe")
	   os.Exit(1)
	}

	/*** コントロール用のAPIサーバ初期化 ***/
	// Echoセットアップ
	e := echo.New()
	// エラー発生時の対処設定
	e.Use(middleware.Recover())

	// ブラウザからjavascriptを使ってAPI呼び出しできるようにCORS対応
	e.Use(middleware.CORSWithConfig(middleware.CORSConfig{
		AllowOrigins: []string{"*"},
		AllowMethods: []string{echo.GET, echo.PUT, echo.POST, echo.DELETE, echo.HEAD},
	}))

	// ストレージ・サーバのリクエストハンドラ登録
	e.GET("/", requestHandler)
	e.POST(conf.Control_url, controlHandler)

	// サーバの起動
	e.Logger.Fatal(e.Start(conf.Host + ":" + strconv.FormatInt(int64(conf.Port), 10)))
	 
}

